from dataclasses import dataclass

from ._impl import ValidationError

__all__ = ["ValidationError", "SchemaValidationError", "ErrorItem"]


@dataclass
class ErrorItem:
    message: str
    instance_path: str
    schema_path: str


class SchemaValidationError(ValidationError):
    def __init__(self, errors: list[ErrorItem]) -> None:
        self.errors = errors



import sys
import uuid
from dataclasses import dataclass
from datetime import datetime, timezone, timedelta, time, date
from decimal import Decimal
from enum import Enum
from zoneinfo import ZoneInfo

import pytest

from serpyco_rs import ValidationError
from serpyco_rs import Serializer


@pytest.mark.parametrize(
    ["type", "value"],
    (
        (int, 44),
        (str, "123"),
        (float, 4.3),
        (bool, True),
    ),
)
def test_simple_types(type, value):
    serializer = Serializer(type)
    assert serializer.load(serializer.dump(value)) == value


def test_decimal():
    serializer = Serializer(Decimal)
    assert serializer.dump(Decimal(123)) == "123"
    assert serializer.load(123) == Decimal(123)
    assert serializer.load("123") == Decimal(123)


def test_decimal_invalid_value__raise_validation_error():
    serializer = Serializer(Decimal)

    with pytest.raises(ValidationError):
        serializer.load("asd")


def test_dict_encoder():
    serializer = Serializer(dict[str, Decimal])
    val = {"a": Decimal("123.3")}
    assert serializer.dump(val) == {"a": "123.3"}
    assert serializer.load({"a": "123.3"}) == val


def test_array_encoder():
    serializer = Serializer(list[int])
    val = [1, 2, 3]
    assert serializer.dump(val) == serializer.load(val) == val


def test_entity_encoder():
    @dataclass
    class A:
        int_f: int
        float_f: float
        bool_f: bool
        str_f: str

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        float_f=3.14,
        bool_f=True,
        str_f="Test",
    )
    expected = {"bool_f": True, "float_f": 3.14, "int_f": 123, "str_f": "Test"}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_uuid():
    serializer = Serializer(uuid.UUID)
    val = uuid.uuid4()
    assert serializer.dump(val) == str(val)
    assert serializer.load(str(val)) == val


def test_enum():
    class Foo(Enum):
        foo = "foo"

    serializer = Serializer(Foo)
    assert serializer.dump(Foo.foo) == "foo"
    assert serializer.load("foo") == Foo.foo


def test_tuple():
    serializer = Serializer(tuple[int, bool, str])
    assert serializer.dump((1, True, "s")) == [1, True, "s"]
    assert serializer.load([1, True, "s"]) == (1, True, "s")


def test_tuple__invalid_number_items():
    serializer = Serializer(tuple[int, bool, str])

    with pytest.raises(ValidationError) as exec_info:
        serializer.dump((1,))
    assert exec_info.value.args[0] == "Invalid number of items for tuple"

    with pytest.raises(ValidationError) as exec_info:
        serializer.load((1,), validate=False)
    assert exec_info.value.args[0] == "Invalid number of items for tuple"


@pytest.mark.parametrize(
    ["value", "expected"],
    (
        (datetime(2022, 10, 10, 14, 23, 43), "2022-10-10T14:23:43"),
        (datetime(2022, 10, 10, 14, 23, 43, 123456), "2022-10-10T14:23:43.123456"),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone.utc),
            "2022-10-10T14:23:43+00:00",
        ),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone(timedelta(hours=1))),
            "2022-10-10T14:23:43+01:00",
        ),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=ZoneInfo("Europe/Berlin")),
            "2022-10-10T14:23:43+02:00",
        ),
    ),
)
def test_datetime_dump(value, expected):
    serializer = Serializer(datetime)
    assert serializer.dump(value) == expected


@pytest.mark.parametrize(
    ["value", "expected"],
    (
        # ('2022-10-10T14:23:43', datetime(2022, 10, 10, 14, 23, 43)),
        # ('2022-10-10T14:23:43.123456', datetime(2022, 10, 10, 14, 23, 43, 123456)),
        (
            "2022-10-10T14:23:43+00:00",
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone.utc),
        ),
        (
            "2022-10-10T14:23:43+01:00",
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone(timedelta(hours=1))),
        ),
        (
            "2022-10-10T14:23:43+02:00",
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=ZoneInfo("Europe/Berlin")),
        ),
    ),
)
def test_datetime_load(value, expected):
    serializer = Serializer(datetime)
    assert serializer.load(value) == expected


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("12:34", time(12, 34)),
        ("12:34:56", time(12, 34, 56)),
        ("12:34:56.000078", time(12, 34, 56, 78)),
        # ('12:34:56.000078+00:00', time(12, 34, 56, 78, tzinfo=tzoffset(None, 0))),
    ],
)
def test_time_load(value, expected):
    serializer = Serializer(time)
    assert serializer.load(value) == expected


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        (time(12, 34), "12:34:00"),
        (time(12, 34, 56), "12:34:56"),
        (time(12, 34, 56, 78), "12:34:56.000078"),
    ],
)
def test_time_dump(value, expected):
    serializer = Serializer(time)
    assert serializer.dump(value) == expected


def test_date():
    serializer = Serializer(date)
    assert serializer.load("2022-10-14") == date(2022, 10, 14)
    assert serializer.dump(date(2022, 10, 13)) == "2022-10-13"


if sys.version_info >= (3, 10):

    def test_optional():
        @dataclass
        class T:
            foo: int | None = None

        serializer = Serializer(T)
        assert serializer.dump(T()) == {"foo": None}
        assert serializer.dump(T(foo=1)) == {"foo": 1}
        assert serializer.load({}) == T()
        assert serializer.load({"foo": 12}) == T(foo=12)



from dataclasses import dataclass
from decimal import Decimal
from typing import Union


@dataclass(frozen=True)
class Min:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class Max:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class MinLength:
    value: int


@dataclass(frozen=True)
class MaxLength:
    value: int


@dataclass(frozen=True)
class Places:
    value: int



from abc import ABCMeta, abstractmethod
from typing import Any

import jsonschema_rs

from serpyco_rs.exceptions import SchemaValidationError, ErrorItem


class Validator(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, schema: dict[str, Any]) -> None:
        ...

    @abstractmethod
    def validate(self, data: Any) -> None:
        ...


class JsonschemaRSValidator(Validator):
    def __init__(self, schema: dict[str, Any]) -> None:
        self._validator = jsonschema_rs.JSONSchema(schema)

    def validate(self, data: Any) -> None:
        if not self._validator.is_valid(data):
            errors = list(self._validator.iter_errors(data))
            if errors:
                raise SchemaValidationError([self._map_err(e) for e in errors])

    def _map_err(self, err: jsonschema_rs.ValidationError) -> ErrorItem:
        return ErrorItem(
            message=err.message,
            instance_path="/".join(map(str, err.instance_path)),
            schema_path="/".join(map(str, err.schema_path)),
        )



import sys
from dataclasses import dataclass, field

import pytest

from serpyco_rs import Serializer, SchemaValidationError
from typing import List, Optional
from collections.abc import Sequence, Mapping


def test_dump_simple_fields_types():
    @dataclass
    class A:
        int_f: int
        float_f: float
        bool_f: bool
        str_f: str

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        float_f=3.14,
        bool_f=True,
        str_f="Test",
    )
    expected = {"bool_f": True, "float_f": 3.14, "int_f": 123, "str_f": "Test"}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_simple_nested_dataclasses():
    @dataclass
    class C:
        value: int

    @dataclass
    class B:
        value: str
        nested: C

    @dataclass
    class A:
        int_f: int
        nested: B

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        nested=B(value="test", nested=C(value=1)),
    )

    expected = {"int_f": 123, "nested": {"nested": {"value": 1}, "value": "test"}}

    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_iterables():
    @dataclass
    class A:
        iterable_builtins_list: list[int]
        iterable_typing_list: List[int]
        iterable_builtins_sequence: Sequence[int]

    serializer = Serializer(A)

    obj = A(
        iterable_builtins_list=[1, 2, 3],
        iterable_typing_list=[1, 2, 3],
        iterable_builtins_sequence=[1, 2, 3],
    )

    expected = {
        "iterable_builtins_list": [1, 2, 3],
        "iterable_typing_list": [1, 2, 3],
        "iterable_builtins_sequence": [1, 2, 3],
    }

    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_mappings():
    @dataclass
    class B:
        value: str

    @dataclass
    class A:
        dict_field: dict[str, int]
        mapping_field: Mapping[str, B]

    serializer = Serializer(A)

    obj = A(dict_field={"foo": 1}, mapping_field={"bar": B(value="123")})

    expected = {
        "dict_field": {"foo": 1},
        "mapping_field": {"bar": {"value": "123"}},
    }

    assert serializer.load(expected) == obj
    assert serializer.dump(obj) == expected


def test_required_and_nullable():
    @dataclass
    class ReqNotNull:
        foo: int

    @dataclass
    class ReqNullable:
        foo: Optional[int]

    @dataclass
    class OptionalNotNull:
        foo: int = 1

    @dataclass
    class OptionalNullable:
        foo: Optional[int] = 1

    req_not_null = Serializer(ReqNotNull)
    req_nullable = Serializer(ReqNullable)
    optional_not_null = Serializer(OptionalNotNull)
    optional_nullable = Serializer(OptionalNullable)

    assert req_not_null.load({"foo": 2}) == ReqNotNull(foo=2)
    with pytest.raises(SchemaValidationError):
        req_not_null.load({"foo": None})
    with pytest.raises(SchemaValidationError):
        req_not_null.load({})

    assert req_nullable.load({"foo": 2}) == ReqNullable(foo=2)
    assert req_nullable.load({"foo": None}) == ReqNullable(foo=None)
    with pytest.raises(SchemaValidationError):
        req_nullable.load({})

    assert optional_not_null.load({"foo": 2}) == OptionalNotNull(foo=2)
    with pytest.raises(SchemaValidationError):
        assert optional_not_null.load({"foo": None})
    assert optional_not_null.load({}) == OptionalNotNull(foo=1)

    assert optional_nullable.load({"foo": 2}) == OptionalNullable(foo=2)
    assert optional_nullable.load({"foo": None}) == OptionalNullable(foo=None)
    assert optional_nullable.load({}) == OptionalNullable(foo=1)


def test_required_and_nullable_list():
    @dataclass
    class Entity:
        foo: Optional[list[Optional[int]]] = None

    entity_serializer = Serializer(Entity)

    assert entity_serializer.load({}) == Entity(foo=None)
    assert entity_serializer.load({"foo": None}) == Entity(foo=None)
    assert entity_serializer.load({"foo": []}) == Entity(foo=[])
    assert entity_serializer.load({"foo": [1]}) == Entity(foo=[1])
    assert entity_serializer.load({"foo": [1, None]}) == Entity(foo=[1, None])


def test_defaults():
    @dataclass
    class Entity:
        foo: str = "123"
        bar: list[int] = field(default_factory=lambda: list([1, 2, 3]))

    entity_serializer = Serializer(Entity)

    assert entity_serializer.load({"bar": [1]}) == Entity(foo="123", bar=[1])
    assert entity_serializer.load({}) == Entity(foo="123", bar=[1, 2, 3])


if sys.version_info >= (3, 10):

    def test_union_optional__dump_load__ok():
        # arrange
        @dataclass
        class UnionClass:
            name: str | None
            count: None | int

        # act
        serializer = Serializer(UnionClass)

        # assert
        foo = UnionClass(name=None, count=None)
        dict_foo = {"name": None, "count": None}
        assert serializer.dump(foo) == dict_foo
        assert foo == serializer.load(dict_foo)

        bar = UnionClass(name="try", count=5)
        dict_bar = {"name": "try", "count": 5}
        assert serializer.dump(bar) == dict_bar
        assert bar == serializer.load(dict_bar)



import dataclasses
import sys
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from decimal import Decimal
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast, get_type_hints
from uuid import UUID
from enum import Enum, IntEnum


from .metadata import Min, Max, MaxLength, MinLength, Places

from attributes_doc import get_attributes_doc


if sys.version_info >= (3, 10):  # pragma: no cover
    from types import UnionType
else:  # pragma: no cover
    UnionType = None

try:
    import attr
except ImportError:  # pragma: no cover
    attr = None  # type: ignore


_NoneType = type(None)

_T = TypeVar("_T")


class NotSet:
    def __repr__(self) -> str:
        return "NOT_SET"


NOT_SET = NotSet()


@dataclasses.dataclass
class Type:
    pass


@dataclasses.dataclass
class IntegerType(Type):
    min: Optional[int] = None
    max: Optional[int] = None


@dataclasses.dataclass
class StringType(Type):
    min_length: Optional[int] = None
    max_length: Optional[int] = None


@dataclasses.dataclass
class BytesType(Type):
    pass


@dataclasses.dataclass
class FloatType(Type):
    min: Optional[float] = None
    max: Optional[float] = None


@dataclasses.dataclass
class DecimalType(Type):
    places: Optional[int] = None
    min: Optional[Decimal] = None
    max: Optional[Decimal] = None


@dataclasses.dataclass
class BooleanType(Type):
    pass


@dataclasses.dataclass
class UUIDType(Type):
    pass


@dataclasses.dataclass
class TimeType(Type):
    pass


@dataclasses.dataclass
class DateTimeType(Type):
    pass


@dataclasses.dataclass
class DateType(Type):
    pass


@dataclasses.dataclass
class EnumType(Type):
    cls: type[Union[Enum, IntEnum]]


@dataclasses.dataclass
class EntityField:
    name: str
    type: Type
    doc: Optional[str] = None
    default: Any = NOT_SET
    default_factory: Union[Callable[[], Any], NotSet] = NOT_SET
    is_property: bool = False


@dataclasses.dataclass
class EntityType(Type):
    cls: type[Any]
    fields: Sequence[EntityField]
    generics: Mapping[TypeVar, Any] = dataclasses.field(default_factory=dict)
    doc: Optional[str] = None
    is_presenter: bool = False


@dataclasses.dataclass
class OptionalType(Type):
    inner: Type


@dataclasses.dataclass
class ArrayType(Type):
    item_type: Type
    is_sequence: bool


@dataclasses.dataclass
class DictionaryType(Type):
    key_type: Type
    value_type: Type
    is_mapping: bool


@dataclasses.dataclass
class TupleType(Type):
    item_types: Sequence[Type]


@dataclasses.dataclass
class AnyType(Type):
    pass


def describe_type(t: Any) -> Type:
    parameters: tuple[Any, ...] = ()
    args: tuple[Any, ...] = ()
    metadata: tuple[Any, ...] = ()
    if hasattr(t, "__origin__"):
        parameters = getattr(t.__origin__, "__parameters__", ())
        args = t.__args__
        metadata = getattr(t, "__metadata__", ())
        t = t.__origin__
    elif UnionType and isinstance(t, UnionType):  # UnionType has no __origin__
        args = t.__args__
        t = Union
    elif hasattr(t, "__parameters__"):
        # Если передан generic-класс без type-параметров, значит по PEP-484 заменяем все параметры на Any
        parameters = t.__parameters__
        args = (Any,) * len(parameters)

    generics = dict(zip(parameters, args))

    if t is Any:
        return AnyType()

    if isinstance(t, type):

        if simple := {
            bytes: BytesType,
            bool: BooleanType,
            date: DateType,
            time: TimeType,
            datetime: DateTimeType,
            UUID: UUIDType,
        }.get(t):
            return simple()

        if number_type := {
            int: IntegerType,
            float: FloatType,
        }.get(t):
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return number_type(
                min=cast(Any, min_meta.value) if min_meta else None,
                max=cast(Any, max_meta.value) if max_meta else None,
            )

        if t is Decimal:
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            places_meta = _find_metadata(metadata, Places)
            return DecimalType(
                min=cast(Decimal, min_meta.value) if min_meta else None,
                max=cast(Decimal, max_meta.value) if max_meta else None,
                places=places_meta.value if places_meta else None,
            )

        if t is str:
            min_length_meta = _find_metadata(metadata, MinLength)
            max_length_meta = _find_metadata(metadata, MaxLength)
            return StringType(
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
            )

        if t in {Sequence, list}:
            return ArrayType(
                item_type=describe_type(args[0]) if args else AnyType(),
                is_sequence=t is Sequence,
            )

        if t in {Mapping, dict}:
            return DictionaryType(
                key_type=describe_type(args[0]) if args else AnyType(),
                value_type=describe_type(args[1]) if args else AnyType(),
                is_mapping=t is Mapping,
            )

        if t is tuple:
            if not args or Ellipsis in args:
                raise RuntimeError("Variable length tuples are not supported")
            return TupleType(item_types=[describe_type(arg) for arg in args])

        if issubclass(t, (Enum, IntEnum)):
            return EnumType(cls=t)

        if dataclasses.is_dataclass(t):
            return _describe_dataclass(t, generics)

        if attr and attr.has(t):
            return _describe_attrs(t, generics)

    if t in {Union}:
        if len(args) != 2 or _NoneType not in args:
            raise RuntimeError(
                f"Only Unions of one type with None are supported: {t}, {args}"
            )
        inner = args[1] if args[0] is _NoneType else args[0]
        return OptionalType(describe_type(inner))

    if isinstance(t, TypeVar):
        raise RuntimeError(f"Unfilled TypeVar: {t}")

    raise RuntimeError(f"Unknown type {t!r}")


if not TYPE_CHECKING:
    # Mypy считает, что типы unhashable
    describe_type = lru_cache()(describe_type)


def _describe_dataclass(t: type[Any], generics: Mapping[TypeVar, Any]) -> EntityType:
    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}

    fields = []
    for field in dataclasses.fields(t):
        fields.append(
            EntityField(
                name=field.name,
                doc=docs.get(field.name),
                type=describe_type(
                    _replace_generics(types.get(field.name, field.type), generics)
                ),
                default=(
                    field.default
                    if field.default is not dataclasses.MISSING
                    else NOT_SET
                ),
                default_factory=(
                    field.default_factory
                    if field.default_factory is not dataclasses.MISSING
                    else NOT_SET
                ),
                is_property=False,
            )
        )

    return EntityType(cls=t, fields=fields, generics=generics, doc=t.__doc__)


def _describe_attrs(t: type[Any], generics: Mapping[TypeVar, Any]) -> EntityType:
    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}
    fields = []
    for field in attr.fields(t):
        default = NOT_SET
        if field.default is not attr.NOTHING and not isinstance(field.default, attr.Factory):  # type: ignore[arg-type]
            default = field.default

        default_factory = NOT_SET
        if isinstance(field.default, attr.Factory):  # type: ignore[arg-type]
            default_factory = field.default.factory

        fields.append(
            EntityField(
                name=field.name,
                doc=docs.get(field.name),
                type=describe_type(
                    _replace_generics(types.get(field.name, field.type), generics)
                ),
                default=default,
                default_factory=default_factory,
                is_property=False,
            )
        )
    return EntityType(cls=t, fields=fields, generics=generics)


def _replace_generics(t: Any, generics: Mapping[TypeVar, Any]) -> Any:
    try:
        if parameters := getattr(t, "__parameters__", None):
            t = t[tuple(generics[parameter] for parameter in parameters)]
        if isinstance(t, TypeVar):
            t = generics[t]
    except KeyError as exc:
        raise RuntimeError(f"Unfilled TypeVar: {exc.args[0]}") from exc
    return t


def _find_metadata(annotations: Iterable[Any], type_: type[_T]) -> Optional[_T]:
    return next((ann for ann in annotations if isinstance(ann, type_)), None)



import sys
from dataclasses import dataclass
from datetime import time, datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Annotated
from uuid import UUID

from serpyco_rs._json_schema import to_json_schema
from serpyco_rs._describe import describe_type
from serpyco_rs.metadata import Min, Max, MinLength, MaxLength


def test_to_json_schema():
    class EnumCls(Enum):
        a = "a"

    @dataclass
    class InnerData:
        """Some important entity"""

        foo: str

    @dataclass
    class Data:
        """Docs"""

        a: Annotated[int, Min(0), Max(10)]
        """A field with bounds"""
        b: float
        c: Decimal
        d: bool
        e: Annotated[str, MinLength(1), MaxLength(5)]
        f: UUID
        g: time
        h: datetime
        i: EnumCls
        j: InnerData
        k: list[int]
        l: tuple[int, str, InnerData]
        m: dict[str, int]
        n: Any

    schema = to_json_schema(describe_type(Data)).dump()

    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "description": "Docs",
        "properties": {
            "a": {
                "description": "A field with bounds",
                "maximum": 10,
                "minimum": 0,
                "type": "integer",
            },
            "b": {"type": "number"},
            "c": {"oneOf": [{"type": "string"}, {"type": "number"}]},
            "d": {"type": "boolean"},
            "e": {"maxLength": 5, "minLength": 1, "type": "string"},
            "f": {"format": "uuid", "type": "string"},
            "g": {
                "format": "regex",
                "pattern": "^[0-9][0-9]:[0-9][0-9](:[0-9][0-9](\\.[0-9]+)?)??(([+-][0-9][0-9]:?[0-9][0-9])|Z)?$",
                "type": "string",
            },
            "h": {
                "format": "regex",
                "pattern": "^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\\.[0-9]+)?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$",
                "type": "string",
            },
            "i": {"enum": ["a"]},
            "j": {
                "description": "Some important entity",
                "properties": {"foo": {"type": "string"}},
                "required": ["foo"],
                "type": "object",
            },
            "k": {"items": {"type": "integer"}, "type": "array"},
            "l": {
                "maxItems": 3,
                "minItems": 3,
                "prefixItems": [
                    {"type": "integer"},
                    {"type": "string"},
                    {
                        "description": "Some important entity",
                        "properties": {"foo": {"type": "string"}},
                        "required": ["foo"],
                        "type": "object",
                    },
                ],
                "type": "array",
            },
            "m": {"additionalProperties": {"type": "integer"}, "type": "object"},
            "n": {},
        },
        "required": [
            "a",
            "b",
            "c",
            "d",
            "e",
            "f",
            "g",
            "h",
            "i",
            "j",
            "k",
            "l",
            "m",
            "n",
        ],
        "type": "object",
    }


if sys.version_info >= (3, 10):

    def test_to_json_schema__new_union_syntax():
        @dataclass
        class Data:
            """Docs"""

            a: int | None

        schema = to_json_schema(describe_type(Data)).dump()

        assert schema == {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "description": "Docs",
            "properties": {
                "a": {"anyOf": [{"type": "null"}, {"type": "integer"}]},
            },
            "required": ["a"],
            "type": "object",
        }



from typing import Any, TypeVar, Generic

from ._describe import describe_type
from ._impl import make_encoder
from ._json_schema import JsonschemaRSValidator, to_json_schema, Validator

T = TypeVar("T", bound=Any)


class Serializer(Generic[T]):
    def __init__(
        self, t: type[T], validator_cls: type[Validator] = JsonschemaRSValidator
    ) -> None:
        type_info = describe_type(t)
        self._encoder = make_encoder(type_info)
        self._validator = validator_cls(to_json_schema(type_info).dump())

    def dump(self, value: T) -> Any:
        return self._encoder.dump(value)

    def load(self, data: Any, validate: bool = True) -> T:
        if validate:
            self._validator.validate(data)
        return self._encoder.load(data)



from __future__ import annotations

from dataclasses import dataclass, asdict
from typing import Any


@dataclass
class Schema:
    title: str | None = None
    description: str | None = None
    default: Any | None = None
    deprecated: bool | None = None
    enum: list[Any] | None = None

    allOf: list[Schema] | None = None
    anyOf: list[Schema] | None = None
    oneOf: list[Schema] | None = None

    def dump(self) -> dict[str, Any]:
        def factory(items: list[tuple[str, Any]]):
            return dict(((k, v) for k, v in items if v is not None))

        schema = asdict(self, dict_factory=factory)
        schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        return schema


@dataclass
class Boolean(Schema):
    type: str = "boolean"


@dataclass
class Null(Schema):
    type: str = "null"


@dataclass
class StringType(Schema):
    type: str = "string"
    minLength: int | None = None
    maxLength: int | None = None
    pattern: str | None = None
    format: str | None = None
    # todo: enum https://json-schema.org/understanding-json-schema/reference/string.html#built-in-formats


@dataclass
class NumberType(Schema):
    type: str = "number"
    multipleOf: int | None = None
    minimum: int | None = None
    exclusiveMinimum: int | None = None
    maximum: int | None = None
    exclusiveMaximum: int | None = None


@dataclass
class IntegerType(NumberType):
    type: str = "integer"


@dataclass
class ObjectType(Schema):
    type: str = "object"
    properties: dict[str, Schema] | None = None
    additionalProperties: bool | Schema | None = None
    required: list[str] | None = None


@dataclass
class ArrayType(Schema):
    type: str = "array"
    items: Schema | None = None
    prefixItems: list[Schema] | None = None
    minItems: int | None = None
    maxItems: int | None = None
    uniqueItems: bool | None = None



import sys
import uuid
from dataclasses import dataclass
from datetime import time, datetime, date
from decimal import Decimal
from enum import Enum
from typing import Annotated, Optional, Any
from unittest import mock

import pytest

from serpyco_rs._describe import describe_type
from serpyco_rs._json_schema import to_json_schema, JsonschemaRSValidator
from serpyco_rs.exceptions import ErrorItem, SchemaValidationError
from serpyco_rs.metadata import MinLength, MaxLength, Min, Max


class EnumTest(Enum):
    foo = "foo"
    bar = "bar"


@dataclass
class EntityTest:
    key: str


@pytest.mark.parametrize(
    ["cls", "value"],
    (
        (bool, True),
        (bool, False),
        (str, ""),
        (Annotated[str, MinLength(1), MaxLength(3)], "12"),
        (int, -99),
        (Annotated[int, Min(1), Max(1000)], 99),
        # (bytes, b'xx'),  # todo: fix bytes validation
        (float, 1.3),
        (Annotated[float, Min(0), Max(0.4)], 0.1),
        (Decimal, "0.1"),  # support str
        (Decimal, 0.1),  # or int input
        (Decimal, "NaN"),  # or int input
        (uuid.UUID, str(uuid.uuid4())),  # support only str input
        (time, "12:34"),
        (time, "12:34Z"),
        (time, "12:34:56"),
        (time, "12:34+0300"),
        (time, "12:34+03:00"),
        (time, "12:34:00+03:00"),
        (time, "12:34:56.000078+03:00"),
        (time, "12:34:56.000078+00:00"),
        # todo: add datetime exemplars
        (datetime, "2022-10-10T14:23:43"),
        (datetime, "2022-10-10T14:23:43.123456"),
        (datetime, "2022-10-10T14:23:43.123456Z"),
        (datetime, "2022-10-10T14:23:43.123456+00:00"),
        (datetime, "2022-10-10T14:23:43.123456-30:00"),
        (date, "2020-07-17"),
        (EnumTest, "foo"),
        (Optional[int], None),
        (Optional[int], 1),
        (EntityTest, {"key": "val"}),
        (list[int], [1, 2]),
        (dict[str, int], {"a": 1}),
        (tuple[str, int, bool], ["1", 2, True]),
        (Any, ["1", 2, True]),
        (Any, {}),
    ),
)
def test_validate(cls, value):
    v = JsonschemaRSValidator(to_json_schema(describe_type(cls)).dump())
    v.validate(value)


if sys.version_info >= (3, 10):

    @pytest.mark.parametrize(
        ["cls", "value"],
        (
            (Optional[int], None),
            (int | None, None),
            (int | None, 2),
        ),
    )
    def test_validate(cls, value):
        v = JsonschemaRSValidator(to_json_schema(describe_type(cls)).dump())
        v.validate(value)


def _mk_e(m=mock.ANY, ip=mock.ANY, sp=mock.ANY) -> ErrorItem:
    return ErrorItem(message=m, instance_path=ip, schema_path=sp)


@pytest.mark.parametrize(
    ["cls", "value", "err"],
    (
        (bool, 1, _mk_e(m='1 is not of type "boolean"')),
        (str, 1, _mk_e(m='1 is not of type "string"')),
        (
            Annotated[str, MinLength(2)],
            "a",
            _mk_e(m='"a" is shorter than 2 characters', sp="minLength"),
        ),
        (
            Annotated[str, MaxLength(2)],
            "aaa",
            _mk_e(m='"aaa" is longer than 2 characters', sp="maxLength"),
        ),
        (int, 9.1, _mk_e(m='9.1 is not of type "integer"')),
        (int, "9", _mk_e(m='"9" is not of type "integer"')),
        (
            Annotated[int, Min(1)],
            0,
            _mk_e(m="0 is less than the minimum of 1", sp="minimum"),
        ),
        (
            Annotated[int, Max(1)],
            10,
            _mk_e(m="10 is greater than the maximum of 1", sp="maximum"),
        ),
        (float, None, _mk_e(m='null is not of type "number"')),
        (
            Annotated[float, Min(1)],
            0.1,
            _mk_e(m="0.1 is less than the minimum of 1", sp="minimum"),
        ),
        (
            Annotated[float, Max(1)],
            10.1,
            _mk_e(m="10.1 is greater than the maximum of 1", sp="maximum"),
        ),
        # (uuid.UUID, "asd", ''),  # todo: validation don't work
        (time, "12:34:a", _mk_e(sp="pattern")),
        (datetime, "2022-10-10//12", _mk_e(sp="pattern")),
        (date, "17-02-2022", _mk_e(sp="pattern")),
        (EnumTest, "buz", _mk_e(m='"buz" is not one of ["foo","bar"]', sp="enum")),
        (
            Optional[int],
            "foo",
            _mk_e(m='"foo" is not valid under any of the given schemas', sp="anyOf"),
        ),
        (EntityTest, {}, _mk_e(m='"key" is a required property', sp="required")),
        (
            list[int],
            [1, "1"],
            _mk_e(m='"1" is not of type "integer"', ip="1", sp="items/type"),
        ),
        (
            dict[str, int],
            {"a": "1"},
            _mk_e(
                m='"1" is not of type "integer"', ip="a", sp="additionalProperties/type"
            ),
        ),
        (
            tuple[str, int, bool],
            ["1"],
            _mk_e(m='["1"] has less than 3 items', sp="minItems"),
        ),
        (
            tuple[str, int, bool],
            ["1", 1, True, 0],
            _mk_e(m='["1",1,true,0] has more than 3 items', sp="maxItems"),
        ),
        # (tuple[str, bool], [1, '1'], ''),   # todo: validation don't work
    ),
)
def test_validate__validation_error(cls, value, err):
    v = JsonschemaRSValidator(to_json_schema(describe_type(cls)).dump())

    with pytest.raises(SchemaValidationError) as exc_info:
        v.validate(value)

    assert exc_info.value.errors == [err]


def test_validate__error_format():
    @dataclass
    class Inner:
        baz: str

    @dataclass
    class A:
        foo: int
        bar: Inner

    v = JsonschemaRSValidator(to_json_schema(describe_type(A)).dump())

    with pytest.raises(SchemaValidationError) as exc_info:
        v.validate({"foo": "1", "bar": {"buz": None}, "qux": 0})

    assert exc_info.value.errors == [
        ErrorItem(
            message='"baz" is a required property',
            instance_path="bar",
            schema_path="properties/bar/required",
        ),
        ErrorItem(
            message='"1" is not of type "integer"',
            instance_path="foo",
            schema_path="properties/foo/type",
        ),
    ]



from ._convert import to_json_schema
from ._entities import (
    Schema,
    Boolean,
    Null,
    StringType,
    NumberType,
    IntegerType,
    ObjectType,
    ArrayType,
)
from ._validate import Validator, JsonschemaRSValidator



import dataclasses
import typing

import pytest
import serpyco
import serpyco_rs


@dataclasses.dataclass
class Nested(object):
    """
    A nested type for Dataclass
    """

    name: str


@dataclasses.dataclass
class Dataclass(object):
    """
    A Dataclass class
    """

    name: str
    value: int
    f: float
    b: bool
    nest: typing.List[Nested]
    many: typing.List[int]
    option: typing.Optional[str] = None


serializer_cython = serpyco.Serializer(Dataclass)
serializer_rs = serpyco_rs.Serializer(Dataclass)

test_object = Dataclass(
    name="Foo",
    value=42,
    f=12.34,
    b=True,
    nest=[Nested(name="Bar_{}".format(index)) for index in range(0, 1000)],
    many=[1, 2, 3],
)


serializers = {
    "cython": serializer_cython,
    "rust": serializer_rs,
}


@pytest.mark.parametrize("impl", ["cython", "rust"])
def test_dump(benchmark, impl):
    serializer = serializers[impl]
    serializer.dump(test_object)  # warmup

    benchmark.group = "dump"
    benchmark.extra_info["impl"] = impl
    benchmark.extra_info["correct"] = (
        serializer.load(serializer.dump(test_object)) == test_object
    )
    benchmark(serializer.dump, test_object)


@pytest.mark.parametrize("impl", ["cython", "rust"])
def test_load(benchmark, impl):
    serializer = serializers[impl]
    test_dict = serializer.dump(test_object)
    serializer.load(test_dict, validate=False)  # warmup

    benchmark.group = "load"
    benchmark.extra_info["impl"] = impl
    benchmark.extra_info["correct"] = (
        serializer.load(serializer.dump(test_object)) == test_object
    )
    benchmark(serializer.load, test_dict, validate=False)


@pytest.mark.parametrize("impl", ["rust", "cython"])
def test_load_validate(benchmark, impl):
    serializer = serializers[impl]
    test_dict = serializer.dump(test_object)
    serializer.load(test_dict, validate=True)  # warmup

    benchmark.group = "load with validate"
    benchmark.extra_info["impl"] = impl
    benchmark.extra_info["correct"] = (
        serializer.load(serializer.dump(test_object)) == test_object
    )

    benchmark(serializer.load, test_dict, validate=True)



from typing import Any, Generic, TypeVar

from ._describe import Type

T = TypeVar("T")

class ValidationError(Exception):
    pass

class Serializer(Generic[T]):
    def dump(self, value: T) -> Any:
        pass
    def load(self, data: Any) -> T:
        pass

def make_encoder(py_class: Type) -> Serializer[T]:
    pass
